;;
;;
;;Table of contents:
;;SECTION 1: Declarations + controls
;;SECTION 1a: Stray class declarations
;;SECTION 2: Tiles
;;SECTION 3: Monsters
;;SECTION 3a: Player
;;SECTION 3b: Hostile monsters
;;SECTION 3c: Monster generation
;;SECTION 4: Items
;;SECTION 4a: Inventory interface
;;SECTION 4b: Weapons
;;SECTION 4c: Armor
;;SECTION 5: Actions
;;SECTION 5a: Combat
;;SECTION 5b: Cheat actions
;;SECTION 6: LOS and Events
;;SECTION 7: Timer and Effects
;;SECTION 8: Stack processing
;;SECTION 9: Generating dungeon
;;SECTION 10: Printing routines
;;SECTION 11: Serialization module
;;SECTION 12: Main sequence

(in-package :cl-user)

(defpackage :sewers
  (:nicknames :r2)
  (:use :common-lisp :curses :md5)
  (:shadow :time :speed :position))

(in-package :sewers)

;;SECTION 1: Declarations + controls

;Constants
(defparameter *dx* 2) (defparameter *dy* 1)
(defparameter *controls* (make-hash-table :test #'eql))
(defparameter *colors* '(:cblue :cdgreen :ccyan :cred :cpurple :cbrown :cgray
                         :cdark :clblue :cgreen :csky :crose :cpink :cyellow
                         :cwhite))
(defparameter *debug-mode* nil)
(defparameter *hiscore-length* 15)

;Meta data
(defparameter *allmaps* (make-array 10 :initial-element nil))
(defparameter *curlevel* 1)
(defparameter *deepest-level* 1)
(defparameter *player* nil)
(defparameter *allmessages* nil)
(defparameter *monster-generator* (make-array 10 :initial-element nil))
(defparameter *item-generator* (make-array 10 :initial-element nil))
(defparameter *timer* nil)

;Map data
(defparameter *map* nil)
(defparameter *known-map* nil)
(defparameter *roompoints* nil)
(defparameter *rooms* nil)
(defparameter *monsterstack* nil
  "Each element is (monster . time)")

;Temporary data
(defparameter *curmonster* nil)
(defparameter *quitflag* nil)
(defparameter *quitreason* "")
(defparameter *messages* nil)
(defparameter *zero-time-action* nil)
(defparameter *cancelled* nil)
(defparameter *new-time* nil)
(defparameter *level-changed* nil)

(define-symbol-macro *upstair* (first *roompoints*))
(define-symbol-macro *downstair* (second *roompoints*))

(defun control-bind (action &rest keys)
  (dolist (i keys)
    (setf (gethash i *controls*) action)))

(defun init-controls ()
  (setf *controls* (make-hash-table :test #'eql))

  ;;Cheat controls
  (when *debug-mode*
    (control-bind (make-instance 'restore-hp) 114 170)
    (control-bind (make-instance 'show-repl-console) 126))

  (if (probe-file "controls.cfg")
    (with-open-file (i "controls.cfg")
      (let* ((*read-eval* nil)
             (remapping (read i)))
        (loop for remap in remapping
             do (apply #'control-bind 
                       (apply #'make-instance (car remap))
                       (cdr remap))))
      nil)
    :error))
    

(defgeneric print-char (obj)
  (:documentation "Returns two values: a character that
  represents tile or monster and it's color (keyword)"))

(defgeneric description (obj)
  (:documentation "Returns a string, describing a tile or monster")
  (:method (x) (declare (ignore x)) "No description yet."))

(defun random-list (list)
  (nth (random (length list)) list))

;;SECTION 1a: Stray class declarations

                                   
(defclass event () 
  ((interesting :initform t :initarg :interesting :accessor interesting)
   (x :initarg :x :initform 0 :accessor x)
   (y :initarg :y :initform 0 :accessor y)))

(defclass e-step (event) 
  ((interesting :initform nil)
   (monster :initarg :monster :accessor monster)))

(defclass e-wait (event) 
  ((interesting :initform nil)
   (monster :initarg :monster :accessor monster)))

(defclass e-combat (event)
  ((attacker :initarg :attacker :accessor attacker)
   (defender :initarg :defender :accessor defender)))

;;SECTION 2: Tiles

(defclass tile () 
  ((x :initarg :x :accessor x) (y :initarg :y :accessor y)
   (monster :initarg :monster :initform nil :accessor monster)
   (passable :initform nil :reader passable)
   (bloodied :initform nil :accessor bloodied)
   (items :initform nil :accessor items)))

(defmethod print-char ((tile tile)) (values #\? :CDARK))

(defmethod print-char :around ((tile tile))
  (if (zerop (bit (fov *player*) (y tile) (x tile)))
      (if (items tile) (print-char (car (items tile)))
              (call-next-method))
      (if (monster tile) (print-char (monster tile))
          (if (items tile) (print-char (car (items tile)))
              (call-next-method)))))  

(defmethod description :around ((tile tile))
  (if (monster tile)
      (format nil "~a ~a" (call-next-method) (description (monster tile)))
      (call-next-method)))

(defclass concrete-floor (tile) ((passable :initform t)))

(defmethod print-char ((tile concrete-floor))
  (values #\. (if (bloodied tile) :CRED :CDARK)))

(defmethod description ((tile concrete-floor))
  "Concrete floor.")

(defclass wall (tile) ())

(defmethod print-char ((tile wall))
  (values #\# (if (bloodied tile) :CRED :CDARK)))

(defmethod description ((tile wall))
  "Concrete wall.")

(defclass staircase (concrete-floor) 
  ((direction :initarg :direction :reader direction)))

(defmethod print-char ((tile staircase)) 
  (values (ecase (direction tile) (:u #\<) (:d #\>))
          (if (bloodied tile) :CRED :CDARK)))

(defmethod description ((tile staircase))
  (format nil "A staircase leading ~a." (ecase (direction tile) (:u "up") (:d "down"))))

(defclass door (tile) ((passable :accessor passable :initarg :passable)))

(defmethod print-char ((tile door))
  (values (if (passable tile) #\/ #\+) :cbrown))

(defmethod description ((tile door))
  (if (passable tile) "An open door." "A closed door."))

;;SECTION 3: Monsters

(defgeneric ai (monster)
  (:documentation "When called, returns an action that monster
  attempts to do."))

(defgeneric do-action (monster action)
  (:documentation "Execute an action"))

(defgeneric kill (monster &optional quitflag quitreason)
  (:documentation "Kill teh monster!"))

(defgeneric react (monster event)
  (:documentation "How a monster would react to an event.")
  (:method (m e) (declare (ignore m e)) nil))

(defclass monster ()
  ((name :initform "monster" :initarg :name :accessor name)
   (x :initform 0 :initarg :x :accessor x) 
   (y :initform 0 :initarg :y :accessor y)
   (hp :initform 0 :initarg :hp :accessor hp) 
   (max-hp :initarg :max-hp :accessor max-hp)
   (inventory :initform nil :accessor inventory)
   (hat :initform nil :accessor hat)
   (shirt :initform nil :accessor shirt)
   (boots :initform nil :accessor boots)
   (weapon :initform nil :accessor weapon)
   (speed :initform 1 :accessor speed :documentation "speed modifier")
   (state :initform nil :accessor state :documentation "AI state")
   (target :initform nil :accessor target)
   (strength :initform 2 :accessor strength)
   (dexterity :initform 2 :accessor dexterity)
   (awareness :initform 4 :accessor awareness)
   (foved-p :initform nil :reader foved-p)
   (cash :initform 0 :accessor cash)
   (kill-hash :initform (make-hash-table :test #'equal) :accessor kill-hash)
   (kill-count :initform 0 :accessor kill-count)
   (effect-hash :initform (make-hash-table) :accessor effect-hash)
   ))

(defmethod initialize-instance :after ((monster monster) &key)
  (setf (hp monster) (max-hp monster)))

(defmethod print-char ((monster monster)) (values #\? :CPINK))

(defmethod kill ((monster monster) &optional quitflag quitreason)
  (declare (ignore quitflag quitreason))
  (cast (make-instance 'e-dead :monster monster))
  (setf *monsterstack* (delete monster *monsterstack* :key #'car))
  (unless (eql monster *player*)
    (dolist (i (inventory monster))
      (do-action monster (make-instance 'drop-item :item i))))
  (setf (monster (aref *map* (y monster) (x monster))) nil)
  (maybe-out-char (aref *map* (y monster) (x monster))))


(defun neighbors (y x)
  (list (cons (1+ y) x) (cons (1- y) x) 
        (cons (1+ y) (1+ x)) (cons (1+ y) (1- x))
        (cons y (1+ x)) (cons y (1- x)) 
        (cons (1- y) (1+ x)) (cons (1- y) (1- x))))

(defun free-neighbor-deltas (y x)
  (loop for n in (remove-if (lambda (a) (occupied (car a) (cdr a)))
                            (neighbors y x))
       collecting (cons (- (car n) y) (- (cdr n) x))))

(defun near (monster1 monster2)
  (and (<= (abs (- (x monster1) (x monster2))) 1)
       (<= (abs (- (y monster1) (y monster2))) 1)
       (> (hp monster1) 0) (> (hp monster2) 0)))

(defun nearby-monsters (monster)
  (loop for tile in (neighbors (y monster) (x monster))
        for m = (monster (aref *map* (car tile) (cdr tile)))
        when m collect m))

(defun find-nearby-monster (monster type)
  (let ((mons (remove-if (complement (lambda (x) (typep x type)))
                         (nearby-monsters monster))))
    (if mons (random-list mons) nil)))

(defun follow (monster1 monster2 &key can-open-doors)
  (let ((x1 (x monster1)) (y1 (y monster1))
        (x2 (x monster2)) (y2 (y monster2))
        (filter (if can-open-doors 
                    (lambda (a) 
                      (and (occupied (car a) (cdr a))
                           (not (typep (aref *map* (car a) (cdr a)) 'door))))
                    (lambda (a)
                      (occupied (car a) (cdr a))))))
    (let ((possible-moves (remove-if filter (neighbors y1 x1))))
      (when possible-moves
          (loop with bestm = (car possible-moves)
             with temp = 1000
             for m in possible-moves
             for d = (+ (distance (car m) (cdr m) y2 x2)
                        (if (occupied (car m) (cdr m)) 1 0))
             when (< d temp) do (setf temp d bestm m)
             finally (return bestm))))))

(defun register-kill (killer victim)
  (let ((n (gethash (name victim) (kill-hash killer))))
    (setf (gethash (name victim) (kill-hash killer))
          (if n (1+ n) 1)))
  (incf (kill-count killer)))

;;SECTION 3a: Player

(defclass player (monster) 
  ((name :initform "you")
   (max-hp :initform 10)
   (speed :initform 1)
   (fov :initform (make-array '(21 76) :element-type 'bit) :accessor fov)
   (foved-p :initform t)
   (thirst :initform 1000 :accessor thirst)
   ))

(defmethod initialize-instance :after ((monster player) &key)
  (start (make-instance 'thirst :monster monster)))
   
(defmethod print-char ((monster player)) (values #\@ :CWHITE))

(defmethod description ((monster player)) "You are here.")

(defmethod ai :before (monster)
  (declare (ignore monster))
  (setf *cancelled* nil *new-time* nil))

(defmethod ai ((monster player))
  ;(format t "sending getch...")
  (move (+ *dy* (y monster)) (+ *dx* (x monster))) (refresh)
  (let ((key (getch)))
    (multiple-value-bind (action present-p) (gethash key *controls*)
      (if present-p 
          (progn (if (zerop (time action)) (setf *zero-time-action* t))
                 action) 
          (make-instance 'no-action :key key)))))

(defmethod ai :before ((monster player))
  (show-status) (show-items (y *player*) (x *player*))
  (if *zero-time-action* (setf *zero-time-action* nil))
  (progn (setf *messages* (nreverse *messages*))
         (merge-messages)
         (if *messages* 
             (display-message (pop *messages*))
             (unless *zero-time-action* (clean-line 0)))))

(defmethod ai :after ((monster player))
  (unless *zero-time-action* (setf *messages* nil)))

(defmethod kill ((monster player) 
                 &optional (quitflag :kill) 
                 (quitreason (if *curmonster* (name *curmonster*) "")))
  (setf *quitflag* quitflag)
  (setf *quitreason* quitreason)
  (call-next-method)
  (show-status)
  (getch))

(defmethod react ((monster player) event)
  (when (interesting event) 
    (message (description event))))

(defmethod react ((monster player) (event e-step))
  (with-slots (monster y x) event
    (when (eql monster *player*)
      (show-items y x))))

;;SECTION 3b: Hostile monsters

(defclass hostile-monster (monster)
  ((state :initform :neutral)
   (cooling-off-time :initarg :cooling-off-time 
                     :initform 50 :accessor cooling-off-time)))

(defmethod ai ((monster hostile-monster))
  (case (state monster)
    (:neutral
     (let ((fnd (free-neighbor-deltas (y monster) (x monster))))
       (if fnd
           (let ((d (random-list fnd)))
             (make-instance 'move-to :dx (cdr d) :dy (car d)))
           (make-instance 'wait))))
    (:hostile
     (when (near monster (target monster))
       (return-from ai (make-instance 'attack :monster (target monster))))
     (when (or (<= (hp (target monster)) 0) (zerop (random (cooling-off-time monster))))
       (setf (state monster) :neutral) (setf (target monster) nil)
       (cast (make-instance 'e-m-calm :monster monster))
       (return-from ai (ai monster)))
     (let ((pos (follow monster (target monster))))
       (if pos
           (destructuring-bind (yy . xx) pos
             (make-instance 'move-to
                            :dx (- xx (x monster))
                            :dy (- yy (y monster))))
           (make-instance 'wait))))))

(defun anger (monster target)
  (unless (eql (state monster) :hostile)
    ;;(format t "~a gets angry.~%" monster)
    (setf (state monster) :hostile (target monster) target)
    (cast (make-instance 'e-m-angry :monster monster))))

(defmethod react ((monster hostile-monster) (event e-step))
  (when (typep (monster event) 'player)
    (anger monster (monster event))))

(defmethod react ((monster hostile-monster) (event e-combat))
  (when (typep (attacker event) 'player)
    (anger monster (attacker event))))

(defmethod react ((monster hostile-monster) (event e-wait))
  (when (typep (monster event) 'player)
    (anger monster (monster event))))
    
(defmacro defmonster (name-class char color print-name description &rest stats)
  (destructuring-bind (int-name . mixins) (if (symbolp name-class)
                                              (cons name-class nil)
                                              name-class)
    `(progn
       (defclass ,int-name (,@mixins hostile-monster)
         ((name :initform ,print-name)
          ,@(loop for s on stats by #'cddr
               collect `(,(car s) :initform ,(cadr s)))))
       (defmethod print-char ((monster ,int-name)) (values ,char ,color))
       (defmethod description ((monster ,int-name)) ,description))))

(defmonster lab-rat #\r :cwhite "lab rat"
            "A lab rat that escaped its horrible fate."
            max-hp 2
            speed 8/10
            strength 1
            dexterity 3
            awareness 4
            cooling-off-time 20
            weapon :teeth)

(defmethod damage ((weapon (eql :teeth))) 1)

(defmethod to-hit ((weapon (eql :teeth))) nil)

(defmethod attack-word ((weapon (eql :teeth)) form)
  (if (eql form :he) "bites" "bite"))

(defmonster rat #\r :cgray "rat"
            "A common rat."
            max-hp 1
            speed 1
            strength 1
            dexterity 2
            awareness 4
            cooling-off-time 30
            weapon :teeth)

(defmonster giant-rat #\r :cbrown "giant rat"
            "A giant rat (of Sumatra)."
            max-hp 3
            speed 1
            strength 2
            dexterity 2
            awareness 4
            cooling-off-time 30
            weapon :teeth)

(defmonster rous #\r :crose "rous"
            "A rodent of unusual size."
            max-hp 6
            speed 11/10
            strength 3
            dexterity 2
            awareness 4
            cooling-off-time 30
            weapon :teeth)

(defmonster bat #\b :cgray "bat"
            "A common bat."
            max-hp 1
            speed 1/2
            strength 1
            dexterity 4
            awareness 5
            cooling-off-time 20
            weapon :teeth)

(defmonster turtle #\t :cdgreen "turtle"
            "Formerly a pet turtle, it is now a vicious predator."
            max-hp 4
            speed 2
            strength 3
            dexterity 1
            cooling-off-time 50
            weapon :teeth
            shirt :turtleshell)

(defmethod armor-dv ((item (eql :turtleshell))) 0)

(defmethod armor-pv ((item (eql :turtleshell))) 5)

(defmonster sewer-alligator #\A :cdgreen "sewer alligator"
            "This alligator was flushed down the toilet when it was young."
            max-hp 8
            speed 3/2
            strength 7
            dexterity 1
            awareness 3
            cooling-off-time 50
            weapon :teeth
            shirt :scales)

(defmethod armor-dv ((item (eql :scales))) 1)

(defmethod armor-pv ((item (eql :scales))) 2)


(defmonster spider #\s :cgray "spider"
            "A common spider."
            max-hp 3
            speed 12/10
            strength 2
            dexterity 2
            awareness 4
            cooling-off-time 30
            weapon :teeth)

(defmethod attack-effect ((monster spider))
  (when (zerop (random 3)) 'spider-poison))

(defmonster snake #\s :cgreen "snake"
            "A mothafucking snake."
            max-hp 4
            speed 9/10
            strength 2
            dexterity 5
            awareness 5
            cooling-off-time 40
            weapon :teeth)

(defmethod attack-effect ((monster snake))
  (when (zerop (random 3)) 'snake-poison))

(defmonster ooze #\o :cgreen "ooze"
            "Carnivorous slime mould."
            max-hp 6
            speed 3/2
            strength 4
            dexterity 3
            awareness 4
            cooling-off-time 40
            weapon :bump)

(defmethod attack-effect ((monster ooze))
  (when (zerop (random 3)) 'acid-attack))

(defmonster tarbaby #\o :ccyan "tarbaby"
            "A huge blob of tar, jumping around the sewers."
            max-hp 6
            speed 1/2
            strength 2
            dexterity 6
            awareness 2
            cooling-off-time 20
            weapon :bump)

(defmethod attack-effect ((monster tarbaby))
  (when (zerop (random 3)) 'tar-splash))

(defmethod damage ((weapon (eql :bump))) 0)

(defmethod to-hit ((weapon (eql :bump))) 2)

(defmethod attack-word ((weapon (eql :bump)) form)
  (if (eql form :he) "bumps into" "bump into"))

(defmonster mutant-dog #\d :cgreen "mutant dog"
            "A mutated dog."
            max-hp 6
            speed 8/10
            strength 3
            dexterity 5
            awareness 6
            weapon :teeth
            cooling-off-time 100)

(defclass humanoid (hostile-monster) ())

(defmethod ai ((monster humanoid))
  (case (state monster)
    (:hostile
     (when (near monster (target monster))
       (return-from ai (make-instance 'attack :monster (target monster))))
     (when (or (<= (hp (target monster)) 0) (zerop (random (cooling-off-time monster))))
       (setf (state monster) :neutral) (setf (target monster) nil)
       (cast (make-instance 'e-m-calm :monster monster))
       (return-from ai (ai monster)))
     (let ((pos (follow monster (target monster) :can-open-doors t)))
       (if pos
           (destructuring-bind (yy . xx) pos
             (if (occupied yy xx)
                 (make-instance 'open-door :door (aref *map* yy xx))
                 (make-instance 'move-to
                                :dx (- xx (x monster))
                                :dy (- yy (y monster)))))
           (make-instance 'wait))))
    (t (call-next-method))))

(defmethod initialize-instance :after ((monster humanoid) &key)
  ;Create various equipment for a monster. Only weapons are generated at this time.
  (when (zerop (random (- 12 *curlevel*)))
    (progn (obtain monster 
                   (case (random 5)
                     ((0 1 2) (make-instance 'knife))
                     ((3 4) (make-instance 'axe))))
           (setf (weapon monster) (car (inventory monster)))))) 
      

(defmonster (slime-creature humanoid) #\S :cgreen "slime creature"
            "A humanoid blob of slime."
            max-hp 15
            speed 11/10
            strength 8
            dexterity 2
            awareness 4
            cooling-off-time 100)

(defmonster (crab-man humanoid) #\C :crose "crab man"
            "A member of the misterious race of crab people."
            max-hp 20
            speed 1
            strength 8
            dexterity 4
            awareness 4
            cooling-off-time 100)

(defmethod kill :after ((monster crab-man) &optional q1 q2)
  (declare (ignore q1 q2))
  (when (zerop (random 3))
    (spawn-item (make-instance 'dollars :amount (1+ (random 20))) 
                :position (cons (y monster) (x monster))
                :no-mods t)))

(defmonster (mutant humanoid) #\@ :cgreen "mutant"
            "A mutated human."
            max-hp 10
            speed 1
            strength 4
            dexterity 4
            awareness 4
            cooling-off-time 100)

;;SECTION 3c: Monster generation

(defparameter *monster-table*
;;(monster rarity min-level pref-level)
  '((rat 80 1 1)
    (lab-rat 20 1 2)
    (spider 30 1 2)
    (giant-rat 60 2 3)
    (turtle 30 2 3)
    (bat 30 1 2)
    (snake 60 3 4)
    (rous 40 3 4)
    (ooze 40 4 5)
    (sewer-alligator 30 3 5) 
    (mutant 70 6 7)
    (mutant-dog 50 5 6)
    (tarbaby 40 6 8)
    (slime-creature 70 7 8)
    (crab-man 60 8 10)
    ))

(defun random-array (array)
  (aref array (random (length array))))

(defun calc-rarity (level rarity min-level pref-level)
  (cond
    ((< level min-level) 0)
    ((= level pref-level) rarity)
    (t (floor (/ rarity (expt (1+ (abs (- level pref-level))) 1.2) 1/2)))))

(defun generate-monster (level)
  (if (aref *monster-generator* (1- level))
      (spawn (make-instance (random-array (aref *monster-generator* (1- level)))))
      (progn
        (setf (aref *monster-generator* (1- level))
              (coerce (loop for (monster rarity min-level pref-level) in *monster-table*
                     for rar = (calc-rarity level rarity min-level pref-level)
                     nconcing (loop for i from 1 to rar collecting monster)) 'vector))
        (spawn (make-instance (random-array (aref *monster-generator* (1- level))))))))
        

(defun generate-monsters (level)
  (loop for x from 1 to (+ 4 level (random 5))
       do (generate-monster level)))

(defun test-level (level)
  (loop with res = (make-hash-table)
        with arr = (aref *monster-generator* (1- level))
        repeat 100
        for m = (random-array arr)
        do (incf (gethash m res 0))
        finally (maphash (lambda (key val)
                           (format t "~a: ~d~%" key val))
                         res)))

(defun refresh-monster-generator ()
  (loop for level from 1 to 10
       do (setf (aref *monster-generator* (1- level))
                (coerce (loop for (monster rarity min-level pref-level) 
                                  in *monster-table*
                           for rar = (calc-rarity level rarity 
                                                  min-level pref-level)
                           nconcing (loop for i from 1 to rar 
                                       collecting monster))
                        'vector))))
       
(defun test-levels ()
  (refresh-monster-generator)
  (let ((h (make-hash-table)))
    (loop for i from 0 to 9 do
         (loop with arr = (aref *monster-generator* i)
               repeat 1000
               for m = (random-array arr)
               for g = (gethash m h)
               unless g do 
                (setf g (make-array 10 :initial-element 0))
                (setf (gethash m h) g)
               do (incf (aref g i))))
    (maphash (lambda (key val)
               (format t "~15a:~{ ~2d~}~%" key 
                       (map 'list (lambda (v) (round v 10)) val))) h)))

;;SECTION 4: Items

(defclass item () 
  ((name :initarg :name :accessor name)
   (x :initarg :x :writer (setf x))
   (y :initarg :y :writer (setf y))
   (parent :initarg :parent :initform nil :accessor parent)))

(defmethod x ((item item))
  (with-slots (x parent) item
    (if parent (x parent) x)))

(defmethod y ((item item))
  (with-slots (y parent) item
    (if parent (y parent) y)))

(defmethod print-char ((item item)) (values #\~ :CROSE))

(defmethod description ((item item)) (name item))

(defun obtain (monster item)
  "Generates item in monster's inventory"
  (setf (parent item) monster)
  (push item (inventory monster)))

(defun destroy (item)
  (with-slots (parent) item
    (if parent
        (progn
          (setf (inventory parent) (delete item (inventory parent)))
          (setf parent nil))
        (let ((tile (aref *map* (y item) (x item))))
          (setf (items tile) (delete item (items tile)))))))

(defun random-modifier ()
  (case (random 50)
    ((1 3 5 7) -1)
    ((0 2 4 6 10 12) 1)
    ((11 16) 2)
    ((20) 3)
    (t 0)))

(defun spawn-item (item &key (position (random-position)) (no-mods nil))
  (destructuring-bind (y . x) position
    (setf (y item) y (x item) x)
    (push item (items (aref *map* y x)))
    (unless no-mods
      (typecase item
        (weapon (setf (damage-mod item) (random-modifier))
                (setf (to-hit-mod item) (random-modifier)))
        (armor (setf (dv-mod item) (random-modifier))
               (setf (pv-mod item) (random-modifier)))
        (money (setf (amount item) (random-money-amount *curlevel*)))))))
       
(defparameter *item-table*
  ;;(item rarity min-level)
  '((t-shirt 10 1)
    (knife 20 1)
    (cola-can 20 1)
    (beer-can 20 1)
    (energy-drink 20 1)
    (juice-can 20 1)
    (sneakers 10 2)
    (baseball-cap 10 1)
    (dollars 100 1)
    (axe 20 4)
    (chainsaw 10 7)))

(defun calc-rarity-items (level rarity min-level)
  (if (< level min-level) 0 rarity))

(defun generate-item (level)
  (if (aref *item-generator* (1- level))
      (spawn-item (make-instance (random-array (aref *item-generator* (1- level)))))
      (progn
        (setf (aref *item-generator* (1- level))
              (coerce (loop for (item rarity min-level) in *item-table*
                     for rar = (calc-rarity-items level rarity min-level)
                     nconcing (loop for i from 1 to rar collecting item)) 'vector))
        (spawn-item (make-instance (random-array (aref *item-generator* 
                                                       (1- level))))))))

(defun generate-items (level)
  (loop for x from 1 to (+ 8 (random 5))
        do (generate-item level)))
;  (spawn-item (make-instance 't-shirt :dv-mod 1 :pv-mod 1))
;  (spawn-item (make-instance 'knife :damage-mod -1))
;  (spawn-item (make-instance 'baseball-cap :dv-mod 1 :pv-mod -1))
;  (spawn-item (make-instance 'sneakers :pv-mod 2)))

(defun equipped (item)
  (let ((mon (parent item)))
    (when (and mon (typep mon 'monster))
      (member item (list (hat mon) (shirt mon) 
                         (boots mon) (weapon mon))))))
    
(defun equippable (item)
  (typecase item
    (weapon 'weapon)
    (hat 'hat)
    (shirt 'shirt)
    (boots 'boots)
    (t nil)))

(defclass usable-item (item) ())

(defgeneric use (monster item)
  (:method (monster item) (declare (ignore monster item)) nil))

(defclass potion (usable-item)
  ((color :initform :CROSE :initarg :color :accessor color)))

(defmethod print-char ((item potion)) (values #\! (color item)))

(defmethod use :after (monster (item potion))
  (declare (ignore monster))
  (destroy item))

(defmethod use :after ((monster player) (item potion))
  (incf (thirst monster) 600)
  (when (> (thirst monster) 1500) (setf (thirst monster) 1500)))

(defclass cola-can (potion)
  ((color :initform :cbrown)
   (name :initform "cola can")))

(defmethod use (monster (item cola-can))
  (setf (hp monster) (min (max-hp monster) (+ 5 (hp monster))))
  (if (eql monster *player*) (message "You drink a refreshing beverage."))
  (when (zerop (random 3))
    (incf (dexterity monster))
    (if (eql monster *player*) (message "You feel more dextrous."))))

(defclass beer-can (potion)
  ((color :initform :cyellow)
   (name :initform "beer can")))

(defmethod use (monster (item beer-can))
  (setf (hp monster) (min (max-hp monster) (+ 5 (hp monster))))
  (if (eql monster *player*) (message "You drink a refreshing beverage."))
  (when (zerop (random 3))
    (incf (strength monster))
    (if (eql monster *player*) (message "You feel more strong."))))

(defclass energy-drink (potion)
  ((color :initform :cpink)
   (name :initform "energy drink")))

(defmethod use (monster (item energy-drink))
  (setf (hp monster) (min (max-hp monster) (+ 5 (hp monster))))
  (if (eql monster *player*) (message "You drink a refreshing beverage."))
  (when (zerop (random 3))
    (incf (awareness monster))
    (if (eql monster *player*) (message "You feel more alert."))))

(defclass juice-can (potion)
  ((color :initform :crose)
   (name :initform "juice can"))) 

(defmethod use (monster (item juice-can) &aux flag)
  (when (zerop (random 3))
    (incf (max-hp monster) 2)
    (setf flag t))
  (setf (hp monster) (min (max-hp monster) (+ 5 (hp monster))))
  (if (eql monster *player*) (message "You drink a refreshing beverage."))
  (when (and flag (eql monster *player*)) (message "You feel more healthy.")))



(defclass money (item)
  ((amount :initarg :amount :initform 0 :accessor amount :accessor cash)))

(defclass dollars (money)
  ((name :initform "$$$")))

(defmethod print-char ((item dollars)) (values #\$ :cdgreen))

(defmethod description ((item dollars))
  (format nil "$~d" (amount item)))

(defun random-money-amount (level)
  (1+ (random (* level 10))))

;;SECTION 4a: Inventory interface

(defclass scrollable-window ()
  ((y :initarg :y :accessor y)
   (x :initarg :x :accessor x)
   (width :initarg :width :accessor width)
   (height :initarg :height :accessor height)
   (content :initarg :content :accessor content
            :documentation "array of items")
   (index :initform 0 :accessor index
          :documentation "Current index")
   (cursor :initform 0 :accessor cursor)
   (caption :initarg :caption :accessor caption)))

(defun current-item (window)
  (aref (content window) (+ (index window) (cursor window))))

(defun scrollable-up (window)
  (> (index window) 0))

(defun scrollable-down (window)
  (> (length (content window)) (+ (index window) (height window))))

(defun scroll-up (window)
  (when (scrollable-up window) 
    (decf (index window))
    (display-window window :refresh-only t)))

(defun scroll-down (window)
  (when (scrollable-down window)
    (incf (index window))
    (display-window window :refresh-only t)))

(defun page-up (window)
  (if (scrollable-up window)
      (with-slots (index height) window
        (setf index (min (max 0 (- index (- height 2))) (1- index)))
        (display-window window :refresh-only t))
      (with-slots (cursor) window
        (attrset :cdark)
        (mvaddch (+ (y window) 1 cursor) (+ (x window) 1) #\Space)
        (setf cursor 0)
        (mvaddch (+ (y window) 1 cursor) (+ (x window) 1) #\>)
        (move (+ (y window) 1 cursor) (+ (x window) 1)))))

(defun page-down (window)
  (if (scrollable-down window)
      (with-slots (index height) window
        (setf index (max (min (- (length (content window))
                                 (height window))
                              (+ index (- height 2))) 
                         (1+ index)))
        (display-window window :refresh-only t))
      (with-slots (cursor) window
        (attrset :cdark)
        (mvaddch (+ (y window) 1 cursor) (+ (x window) 1) #\Space)
        (setf cursor (min (1- (height window)) 
                          (1- (length (content window)))))
        (mvaddch (+ (y window) 1 cursor) (+ (x window) 1) #\>)
        (move (+ (y window) 1 cursor) (+ (x window) 1)))))

(defun move-up (window)
  (with-slots ((wy y) (wx x) cursor) window
    (if (= cursor 0)
        (scroll-up window)
        (progn
          (attrset :cdark)
          (mvaddch (+ wy 1 cursor) (+ wx 1) #\Space)
          (decf cursor)
          (mvaddch (+ wy 1 cursor) (+ wx 1) #\>)
          (move (+ wy 1 cursor) (+ wx 1))))))

(defun move-down (window)
  (with-slots ((wy y) (wx x) cursor) window
    (when (= (+ (index window) (cursor window)) (1- (length (content window))))
      (return-from move-down))
    (if (= cursor (1- (height window)))
        (scroll-down window)
        (progn
          (attrset :cdark)
          (mvaddch (+ wy 1 cursor) (+ wx 1) #\Space)
          (incf cursor)
          (mvaddch (+ wy 1 cursor) (+ wx 1) #\>)
          (move (+ wy 1 cursor) (+ wx 1))))))
         
(defun display-window (window &key refresh-only)
  (with-slots ((wy y) (wx x) width height content index cursor) window
    (attrset :cdark)
    (if refresh-only
        (clean-area (1+ wy) (1+ wx) (+ wy height) (+ wx width))
        (clean-area wy wx (+ wy height 1) (+ wx width 1)))
;;Draw border
    (unless refresh-only
    (mvaddch wy wx #\,)
    (loop for x from (1+ wx) to (+ wx width)
          do (mvaddch wy x #\-)
             (mvaddch (+ wy height 1) x #\-))
    (mvaddch wy (+ wx width 1) #\.)
    (loop with x = (+ wx width 1)
          for y from (1+ wy) to (+ wy height)
          do (mvaddch y wx #\|)
             (mvaddch y x #\|))
    (mvaddch (+ wy height 1) (+ wx width 1) #\')
    (mvaddch (+ wy height 1) wx #\`))
;;Scrollers
    (mvaddch (1+ wy) (+ wx width 1)
             (if (scrollable-up window) #\^ #\|))
    (mvaddch (+ wy height) (+ wx width 1)
             (if (scrollable-down window) #\v #\|))
;;Draw caption
    (attrset :cyellow)
    (mvprintw wy (+ wx 4) (caption window))
;;Draw items
    (loop for y from (1+ wy) to (+ wy height)
          for n from index to (1- (length content))
          for item = (aref content n)
          do (attrset (nth-value 1 (print-char item)))
             (mvprintw y (+ wx 2) (if (equipped item)
                                      (format nil "~a [equipped]"
                                              (description item))
                                      (description item))))
;;Draw cursor
    (attrset :cdark)
    (mvaddch (+ wy 1 cursor) (+ wx 1) #\>)
    (move (+ wy 1 cursor) (+ wx 1))))
  
(defun choose-item (items caption)
 (let ((window (make-instance 'scrollable-window :y 3 :x 10 :width 40
                              :height (max (min (length items) 15) 5)
                              :content (coerce items 'vector)
                              :caption caption)))
   (display-window window)
   (prog1
       (loop (case (getch)
               ((259 56 450 464) (move-up window))
               ((258 50 456 465) (move-down window))
               ((339 458) (page-up window))
               ((338 463) (page-down window))
               ((27 32 122 239) (return))
               (t (return (current-item window)))))
     (redraw-screen))))

(defun display-items (items caption)
  (let ((window (make-instance 'scrollable-window :y 3 :x 10 :width 40
                               :height (max (min (length items) 15) 5)
                               :content (coerce items 'vector)
                               :caption caption)))
    (display-window window)
    (loop (case (getch)
               ((259 56 450 464) (move-up window))
               ((258 50 456 465) (move-down window))
               ((339 458) (page-up window))
               ((338 463) (page-down window))
               ;((27 32 122 239) (return))
               (t (return (current-item window)))))
    (redraw-screen)))

;;SECTION 4b: Weapons

(defclass weapon (item) 
  ((damage-mod :initform 0 :initarg :damage-mod :accessor damage-mod)
   (to-hit-mod :initform 0 :initarg :to-hit-mod :accessor to-hit-mod)))

(defmethod damage :around ((item weapon))
  (+ (damage-mod item) (call-next-method)))

(defmethod to-hit :around ((item weapon))
  (+ (to-hit-mod item) (call-next-method)))

(defmethod description ((item weapon))
  (with-slots (damage-mod to-hit-mod name) item
    (cond
      ((= damage-mod to-hit-mod 0) (name item))
      ((= damage-mod to-hit-mod)
       (format nil "~a ~@d" name damage-mod))
      (t (format nil "~a ~@d ~@d" name to-hit-mod damage-mod)))))


(defmacro defweapon (int-name color print-name 
                       damage-form to-hit-form &optional attack-word-form)
    `(progn
       (defclass ,int-name (weapon)
         ((name :initform ,print-name)))
       (defmethod print-char ((item ,int-name)) (values #\( ,color))
       (defmethod damage ((item ,int-name)) ,damage-form)
       (defmethod to-hit ((item ,int-name)) ,to-hit-form)
       ,(when attack-word-form
              `(defmethod attack-word ((weapon ,int-name) form)
                 ,attack-word-form))))

(defweapon knife :cdark "knife"
           (random 3) 1 (if (eql form :he) "stabs" "stab"))

(defweapon axe :csky "axe"
           (random 6) 0 (if (eql form :he) "hacks at" "hack at"))

(defweapon chainsaw :cred "chainsaw"
           (+ 3 (random 6)) 0)

;;SECTION 4c: Armor

(defgeneric armor-dv (x)
  (:documentation "Returns a DV value of a monster or an armor item"))

(defgeneric armor-pv (x)
  (:documentation "Returns a PV value of a monster or an armor item"))

(defclass armor (item)
  ((dv :initform 0 :initarg :dv :accessor dv)
   (pv :initform 0 :initarg :pv :accessor pv)
   (dv-mod :initform 0 :initarg :dv-mod :accessor dv-mod)
   (pv-mod :initform 0 :initarg :pv-mod :accessor pv-mod)
   (color :initform :crose :initarg :color :accessor color)))

(defmethod armor-dv ((item armor))
  (+ (dv item) (dv-mod item)))

(defmethod armor-pv ((item armor))
  (+ (pv item) (pv-mod item)))

(defmethod description ((item armor))
  (with-slots (dv-mod pv-mod name) item
    (cond 
      ((= dv-mod pv-mod 0) (name item))
      ((= dv-mod pv-mod)
       (format nil "~a ~@d" name dv-mod))
      (t (format nil "~a ~@d ~@d" name dv-mod pv-mod)))))

(defmethod print-char ((item armor)) (values #\[ (color item)))

(defclass shirt (armor) ())

(defclass t-shirt (shirt) ((name :initform "t-shirt")
                           (dv :initform 1) (pv :initform 0)
                           (color :initform (random-list *colors*))))

(defclass hat (armor) ())

(defmethod print-char ((item hat)) (values #\^ (color item)))

(defclass baseball-cap (hat) ((name :initform "baseball cap")
                              (dv :initform 1) (pv :initform 0)
                              (color :initform (random-list *colors*))))

(defclass boots (armor) 
  ((speed-mod :initform 1 :initarg :speed-mod :accessor speed-mod)))

(defmethod print-char ((item boots)) (values #\= (color item)))

(defclass sneakers (boots) ((name :initform "sneakers")
                            (dv :initform 0) (pv :initform 0)
                            (speed-mod :initform 19/20)
                            (color :initform :cwhite)))



;;SECTION 5: Actions

(defclass action ()
  ((time :initform 5 :initarg :time :accessor time)))

(defun instead-do-action (act1 monster act2)
  (declare (ignore act1))
  (do-action monster act2)
  (setf *new-time* (time act2))
  ;;(format t "Did ~a instead. *new-time*=~a.~%" act2 *new-time*)
  )

(defun cancel-action (action &optional message)
  (declare (ignore action))
  (when (and message (eql *curmonster* *player*))
    (message message))
  (setf *cancelled* t *zero-time-action* t))

(defclass move-to (action)
  ((dx :initarg :dx :accessor dx)
   (dy :initarg :dy :accessor dy)))

(defun move-time (dy dx)
  (if (or (> (abs dy) 1) (> (abs dx) 1) 0)
      (if (= 2 (+ (abs dx) (abs dy))) 4 3)))

(defmethod initialize-instance :after ((self move-to) &key)
  (setf (time self) (move-time (dy self) (dx self))))
                    
(defmethod do-action (monster (move-to move-to))
  (with-slots (y x) monster
    (with-slots (dy dx) move-to
      (if (occupied (+ y dy) (+ x dx))
          (cancel-action move-to "Can't move here.")
          (progn
            (setf (monster (aref *map* y x)) nil)
            (maybe-out-char (aref *map* y x))
            (incf y dy) (incf x dx)
            (setf (monster (aref *map* y x)) monster)
            (maybe-out-char (aref *map* y x))
            ;;(setf (time move-to) (move-time (dy move-to) (dx move-to)))
            (if (boots monster) 
                (setf *new-time* (* (time move-to) 
                                        (speed-mod (boots monster)))))
            (if (eql monster *player*)
                (cast (make-instance 'e-step :monster monster :y y :x x)))
            t)))))

(defmethod do-action ((monster player) (move-to move-to))
  (with-slots (y x) monster
    (with-slots (dy dx) move-to
      (when (or (> (abs dy) 1) (> (abs dx) 1)) (return-from do-action nil))
      (let* ((tile (aref *map* (+ y dy) (+ x dx)))
             (monster? (monster tile)))
        (if (and (typep tile 'door) (not (passable tile)))
            (instead-do-action move-to monster
                               (make-instance 'open-door :door tile))
            (if monster? 
                (if t ;(or (eql (state monster?) :angry)
                      ;(yes-or-no "Do you really want to attack?"))
                    (instead-do-action move-to monster 
                                       (make-instance 'attack :monster monster?))
                    (cancel-action move-to))
                (call-next-method)))))))

(defmethod do-action :after ((monster player) (move-to move-to))
  (unless *cancelled*
    (update-fov monster)))

(defclass wait (action) ((time :initform 5)))

(defmethod do-action (monster (action wait)) 
  (declare (ignore action))
  (if (eql monster *player*)
            (cast (make-instance 'e-wait :monster monster 
                                 :y (y monster) :x (x monster))))
  nil)

(defclass quit-game (action) ((time :initform 0)))

(defmethod do-action ((monster player) (quit quit-game))
  (if (yes-or-no "Do you really want to quit without saving?")
      (setf *quitflag* :quit)
      (cancel-action quit)))

(defclass no-action (action)
  ((key :initarg :key :initform 0 :accessor key)
   (time :initform 0)))

(defmethod do-action ((monster player) (action no-action))
  (setf *zero-time-action* nil)
  (message "No action bound to key: ~a" (key action)))

(defclass cancel (wait) ((time :initform 0)))

(defclass next-message (action) ((time :initform 0)))

(defmethod do-action ((monster player) (action next-message))
  (setf *zero-time-action* t)
  (unless *messages* (message "No more messages."))
  ;(format t "~a~%" *messages*)
  ;(display-message (pop *messages*))
  )

(defclass recap (action) ((time :initform 0)))

(defmethod do-action ((monster player) (action recap))
  (recap-messages)
  (getch)
  (redraw-screen))

(defclass look-around (action) ((time :initform 0)))

(defmethod do-action ((monster player) (action look-around))
  (display-simple-message "Look mode. Navigate with [1-9] or arrow keys. Exit with SPACE or ESC.")
  (let ((xx (x monster)) (yy (y monster)))
    (move (+ yy *dy*) (+ xx *dx*)) (refresh)
    (loop (case (getch)
            ((261 54 454) (incf xx))
            ((260 52 452) (decf xx))
            ((259 56 450) (decf yy))
            ((258 50 456) (incf yy))
            ((51 457) (incf xx) (incf yy))
            ((49 455) (decf xx) (incf yy))
            ((57 451) (incf xx) (decf yy))
            ((55 449) (decf xx) (decf yy))
            ((27 113 169 108 164 13 32) (return)))
       (when (< xx 0) (setf xx 0))
       (when (< yy 0) (setf yy 0))
       (when (> xx 75) (setf xx 75))
       (when (> yy 20) (setf yy 20))
       (if (zerop (bit (fov *player*) yy xx))
           (display-simple-message "You can't see this place.")
         (progn (display-simple-message (description (aref *map* yy xx)))
                (show-items yy xx)))
       (move (+ yy *dy*) (+ xx *dx*)) (refresh))
    (show-items yy xx)
    (move (+ yy *dy*) (+ xx *dx*)) (refresh)))

(defclass open-door (action) ((door :initarg :door :accessor door)
                              (time :initform 5)))

(defmethod do-action :after (monster (action open-door))
  (cast (make-instance 'e-open-door :monster monster :door (door action))))

(defmethod do-action (monster (action open-door))
  (declare (ignore monster))
  (setf (passable (door action)) t))

(defmethod do-action :after ((monster player) (action open-door))
  (update-fov monster))

(defclass close-door (action) ((door :initarg :door :accessor door)
                               (time :initform 5)))

(defmethod do-action :after (monster (action close-door))
  (cast (make-instance 'e-close-door :monster monster :door (door action))))

(defmethod do-action (monster (action close-door))
  (declare (ignore monster))
  (if (occupied (y (door action)) (x (door action)))
      (cancel-action action "The door is blocked.")
      (setf (passable (door action)) nil)))

(defmethod do-action :after ((monster player) (action close-door))
  (update-fov monster))

(defclass open-door-smart (action) ())

(defmethod do-action ((monster player) (action open-door-smart))
  (loop with ndoors = 0 and lastdoor 
        for (y . x) in (neighbors (y monster) (x monster))
        for tile = (aref *map* y x)
        if (and (typep tile 'door) (not (passable tile)))
           do (incf ndoors) (setf lastdoor tile)
        finally
          (if (= ndoors 1)
              (instead-do-action action monster 
                                 (make-instance 'open-door :door lastdoor))
              (let ((dir (choose-direction)))
                (if dir
                    (let ((tile (aref *map* (+ (y monster) (car dir))
                                      (+ (x monster) (cdr dir)))))
                      (if (and (typep tile 'door) (not (passable tile)))
                          (instead-do-action action monster
                                             (make-instance 'open-door :door tile))
                          (cancel-action action)))
                    (cancel-action action))))))
                                             
(defclass close-door-smart (action) ())

(defmethod do-action ((monster player) (action close-door-smart))
  (loop with ndoors = 0 and lastdoor 
        for (y . x) in (neighbors (y monster) (x monster))
        for tile = (aref *map* y x)
        if (and (typep tile 'door) (passable tile))
           do (incf ndoors) (setf lastdoor tile)
        finally
          (if (= ndoors 1)
              (instead-do-action action monster 
                                 (make-instance 'close-door :door lastdoor))
              (let ((dir (choose-direction)))
                (if dir
                    (let ((tile (aref *map* (+ (y monster) (car dir)) 
                                      (+ (x monster) (cdr dir)))))
                      (if (and (typep tile 'door) (passable tile))
                          (instead-do-action action monster
                                             (make-instance 'close-door :door tile))
                          (cancel-action action)))
                    (cancel-action action))))))

(defclass save-game (action) ((time :initform 0)))

(defmethod do-action ((monster player) (action save-game))
  (if (yes-or-no "Do you want to save the game and quit?")
      (progn (save-game "current.sav")
             (setf *quitflag* :saved))
      (cancel-action action)))
      
(defclass descend (action) ())

(defmethod do-action ((monster player) (action descend))
  (let ((tile (aref *map* (y monster) (x monster))))
    (if (and (typep tile 'staircase) (eql (direction tile) :d))
        (go-to-level (1+ *curlevel*) :player :u)
        (cancel-action action "Can't descend here."))))

(defclass ascend (action) ())

(defmethod do-action ((monster player) (action ascend))
  (let ((tile (aref *map* (y monster) (x monster))))
    (if (and (typep tile 'staircase) (eql (direction tile) :u))
        (if (> *curlevel* 1)
            (go-to-level (1- *curlevel*) :player :d)
            (if (yes-or-no "Do you want to leave the sewers?")
                (setf *quitflag* :left *curlevel* 0)
                (cancel-action action)))
        (cancel-action action "Can't ascend here."))))

(defclass get-item (action) ((item :initarg :item :accessor item)))

(defmethod do-action (monster (action get-item))
  (destroy (item action))
  (if (typep (item action) 'money)
      (incf (cash monster) (cash (item action)))
      (obtain monster (item action))))

(defclass get-item-on-spot (action) ())

(defmethod do-action ((monster player) (action get-item-on-spot))
  (let ((items (items (aref *map* (y monster) (x monster)))))
    (if items
        (if (cdr items)
            (let ((item (choose-item items "Pick up stuff")))
              (if item
                  (instead-do-action action monster 
                                     (make-instance 'get-item :item item))
                  (cancel-action action)))
            (instead-do-action action monster 
                               (make-instance 'get-item :item (car items))))
        (cancel-action action "No items here."))))
              
(defclass show-inventory (action) ((time :initform 0)))

(defmethod do-action ((monster player) (action show-inventory))
  (if (inventory monster)
      (display-items (inventory monster) "Inventory")
      (message "You have nothing.")))

(defclass drop-item (action) ((item :initarg :item :accessor item)))

(defmethod do-action (monster (action drop-item))
  (let ((item (item action)))
    (destroy item)
    (setf (y item) (y monster) (x item) (x monster))
    (push item (items (aref *map* (y monster) (x monster))))))

(defclass drop-item-menu (action) ())

(defmethod do-action ((monster player) (action drop-item-menu))
  (if (inventory monster)
      (let ((item (choose-item (inventory monster) "Drop stuff")))
        (if item
            (if (equipped item) 
                (cancel-action action "You can't drop an equipped item.")
                (instead-do-action action monster
                                   (make-instance 'drop-item :item item)))
            (cancel-action action)))
      (cancel-action action "You have nothing to drop.")))

(defclass equip-item (action) ((item :initarg :item :accessor item)
                               (place :initarg :place :accessor place)))
  
(defmethod do-action (monster (action equip-item))
  (setf (slot-value monster (place action)) (item action)))

(defclass unequip-item (action) ((item :initarg :item :accessor item)))

(defmethod do-action (monster (action unequip-item))
  (setf (slot-value monster (equippable (item action))) nil))

(defclass show-equipment (action) ((time :initform 0)))

(defmethod do-action ((monster player) (action show-equipment))
  (setf *zero-time-action* nil)
  (let ((items (remove-if-not #'equippable (inventory monster))))
    (if items
        (let ((item (choose-item items "Equip/Unequip")))
          (if item
              (let ((place (equippable item)))
                (if (equipped item)
                    (instead-do-action action monster
                                       (make-instance 'unequip-item
                                                      :item item))
                    (let ((worn (funcall place monster)))
                      (if worn
                          (let ((action1 (make-instance 'unequip-item
                                                        :item worn))
                                (action2 (make-instance 'equip-item
                                                        :item item
                                                        :place place)))
                            (do-action monster action1)
                            (do-action monster action2)
                            (setf *new-time* (+ (time action1)
                                                   (time action2))))
                          (instead-do-action action monster
                                             (make-instance 'equip-item
                                                            :item item
                                                            :place place))))))
              (cancel-action action)))
        (cancel-action action "You have nothing to equip."))))
                
(defclass use-item (action) ((item :initarg :item :accessor item)))

(defmethod do-action (monster (action use-item))
  (use monster (item action)))

(defclass use-item-menu (action) ())

(defmethod do-action ((monster player) (action use-item-menu))
  (let ((items (remove-if-not (lambda (i) (typep i 'usable-item))
                              (inventory monster))))
    (if items
        (let ((item (choose-item items "Use items")))
          (if item
              (instead-do-action action monster
                                 (make-instance 'use-item
                                                :item item))
              (cancel-action action)))
        (cancel-action action "You have no usable items."))))

;;SECTION 5a: Combat

(defclass attack (action)
  ((monster :initarg :monster :accessor monster)
   (time :initform 6)))

(defgeneric damage (x)
  (:documentation "Returns a damage that a monster or a weapon deals"))

(defmethod damage ((monster monster))
  (if (weapon monster) (+ (damage (weapon monster)) 
                          (random (strength monster)))
      (1+ (random (strength monster)))))

(defgeneric to-hit (x)
  (:documentation "Returns a to-hit value of a monster or a weapon"))

(defmethod to-hit ((monster monster))
  (let* ((unarmed-bonus (ceiling (dexterity monster) 2))
         (th (+ (if (weapon monster)
                   (let ((th (to-hit (weapon monster))))
                     (if th th unarmed-bonus))
                   unarmed-bonus) 
                (dexterity monster))))
    (if (> th 0) th 0)))

(defun random0 (n)
  (if (> n 0) (random n) 0))

(defmethod armor-dv ((monster monster))
  (let ((dv (+ (if (shirt monster) (armor-dv (shirt monster)) 0)
               (if (boots monster) (armor-dv (boots monster)) 0)
               (if (hat monster) (armor-dv (hat monster)) 0)
               (dexterity monster))))
    (if (> dv 0) dv 0)))

(defmethod armor-pv ((monster monster))
  (let ((pv (+ (if (shirt monster) (armor-pv (shirt monster)) 0)
               (if (boots monster) (armor-pv (boots monster)) 0)
               (if (hat monster) (armor-pv (hat monster)) 0)
               (strength monster))))
    (if (> pv 0) pv 0)))

(defgeneric attack-effect (x)
  (:documentation "Returns a symbol of effect, or nil if no effect.")
  (:method (x) (declare (ignore x)) nil))

(defmethod attack-effect ((monster monster))
  (attack-effect (weapon monster)))

(defun hit-p (mon1 mon2)
  (when (> (random0 (1+ (to-hit mon1))) (random0 (1+ (armor-dv mon2)))) t))

(defun hit (mon1 mon2)
  (let ((dam (damage mon1))
        (prev (random0 (armor-pv mon2))))
    (if (> dam prev) 
        (progn (decf (hp mon2) (- dam prev)) (- dam prev))
        0)))

(defmethod do-action (monster (attack attack))
  (with-slots ((monster2 monster)) attack
    ;;(format t "~a attacks ~a. ~%" monster monster2)
    (if (hit-p monster monster2)
        (let ((dam (hit monster monster2)))
          (if (> dam 0) (progn
                          (cast (make-instance 'e-c-hit 
                                               :attacker monster
                                               :defender monster2
                                               :damage dam))
                          (let ((eff (attack-effect monster)))
                            (when eff
                              (start (make-instance eff :monster monster2 
                                                    :source monster))))
                          (loop for (y . x) in (neighbors (y monster2) 
                                                          (x monster2))
                                when (zerop (random 50))
                                     do (setf (bloodied (aref *map* y x)) t)
                                        (maybe-out-char (aref *map* y x)))
                          (when (zerop (random 3))
                            (setf (bloodied (aref *map* (y monster2)
                                                  (x monster2))) t)
                            (maybe-out-char (aref *map* (y monster2)
                                                  (x monster2)))))
                                
              (cast (make-instance 'e-c-fail
                                   :attacker monster
                                   :defender monster2)))
          (when (<= (hp monster2) 0) 
            (kill monster2)
            (register-kill monster monster2)))
        (cast (make-instance 'e-c-miss
                             :attacker monster
                             :defender monster2)))))

;;SECTION 5b: Cheat actions

(defclass cheat-action (action) ((time :initform 0)))

(defmethod do-action :around ((monster player) (action cheat-action))
  (when *debug-mode*
    (call-next-method)))

(defclass restore-hp (cheat-action) ())

(defmethod do-action ((monster player) (action restore-hp))
  (setf (hp monster) (max-hp monster)))

(defclass show-repl-console (cheat-action) ())

(defmethod do-action ((monster player) (action show-repl-console))
  (repl-console 15)
  (redraw-screen)) 

;;SECTION 6: LOS and Events

(defun distance (y1 x1 y2 x2)
  (let ((dy (abs (- y1 y2)))
        (dx (abs (- x1 x2))))
    (/ (+ (* 4 (min dx dy)) (* 3 (abs (- dx dy)))) 3)))

(defun tracer (y1 x1 y2 x2)
  (if (= x1 x2) (loop for y from (min y1 y2) to (max y1 y2)
                   collecting (cons y x1))
      (flet ((line (x)
               (/ (+ (* (- x x1 1/2) (+ y2 1/2))
                     (* (- x2 -1/2 x) (+ y1 1/2)))
                  (- x2 x1))))
        (let ((xmin (min (+ x1 1/2) (+ x2 1/2)))
              (xmax (max (+ x1 1/2) (+ x2 1/2))))
          (loop for x from xmin below xmax by 1/2
             for y* = (line x)
             for y** = (line (+ x 1/2))
             for ymin = (floor (min y* y**))
             for ymax = (if (= y* y**) (+ ymin 1) (max y* y**))
             nconcing (let ((l (loop for y from ymin below ymax
                                    collecting (cons y (floor x)))))
                        (if (< y** y*) (nreverse l) l))
                       into pts
             finally (return (remove-duplicates pts :test #'equal)))))))

(defun tracer* (y1 x1 y2 x2)
  "Returns list of coordinates in the reverse order"
  (let ((tr (tracer y1 x1 y2 x2))) 
    (if (equal (car tr) (cons y1 x1)) (nreverse tr) tr)))

(defun can-see-p (monster y x)  
  (loop for (yy . xx) in (cdr (tracer* (y monster) (x monster) y x))
        always (passable (aref *map* yy xx))))

(defun aware-p (monster y x)
  "Whether the monster is aware of the point (x,y)"
  (if (foved-p monster) (= (bit (fov monster) y x) 1)
      (and (<= (distance (y monster) (x monster) y x) (awareness monster))
           (can-see-p monster y x))))

(defun update-fov (monster)
  ;(format t "Updating fov~%")
  (with-slots (fov awareness (my y) (mx x)) monster
    (let ((y1 (max 0 (- my awareness 1)))
          (x1 (max 0 (- mx awareness 1)))
          (y2 (min 20 (+ my awareness 1)))
          (x2 (min 75 (+ mx awareness 1))))
        (loop for y from y1 to y2
              do (loop for x from x1 to x2
                       do (setf (bit fov y x) 
                                (if (and (<= (distance my mx y x)
                                             (awareness monster))
                                         (can-see-p monster y x)) 1 0))
                          (when (= (bit fov y x) 1)
                            (setf (bit *known-map* y x) 1))
                          (out-char (aref *map* y x)))))))
                                   
(defclass event () 
  ((interesting :initform t :initarg :interesting :accessor interesting)
   (x :initarg :x :initform 0 :accessor x)
   (y :initarg :y :initform 0 :accessor y)))

(defgeneric cast (event)
  (:documentation "Make all monsters that feel the event react to it."))

(defmethod cast :before (event)
  (declare (ignorable event))
  ;(format t "Event ~A casted.~%" event)
  (push (cons *curmonster* 0) *monsterstack*))

(defmethod cast (event)
  (loop for m in (remove-duplicates *monsterstack* :key #'car)
        for mon = (car m)
        when (aware-p mon (y event) (x event))
        do (react mon event)))

(defmethod cast :after (event)
  (declare (ignore event))
  (pop *monsterstack*))

(defclass e-step (event) 
  ((interesting :initform nil)
   (monster :initarg :monster :accessor monster)))

(defclass e-wait (event) 
  ((interesting :initform nil)
   (monster :initarg :monster :accessor monster)))

(defclass e-combat (event)
  ((attacker :initarg :attacker :accessor attacker)
   (defender :initarg :defender :accessor defender)))

(defmethod cast ((event e-combat))
  (with-slots (attacker defender) event
    (loop for m in (remove-duplicates *monsterstack* :key #'car)
       for mon = (car m)
       when (or (aware-p mon (y attacker) (x attacker))
                (aware-p mon (y defender) (x defender)))
       do (react mon event))))

(defgeneric combat-situation (attacker defender result)
  (:documentation "Describe a combat situation"))

(defgeneric attack-word (weapon form)
  (:documentation "Returns a verb that would be used in a combat message")
  (:method (weapon form) (declare (ignore weapon))
           (if (eql form :he) "hits" "hit")))

(defmethod attack-word ((weapon (eql nil)) form)
  (if (eql form :he) "punches" "punch"))

(defmethod combat-situation (att def result)
  (format nil "~@(~a~) ~a ~a~a." 
                 (name att) 
                 (if (eql result :hit) 
                     (attack-word (weapon att) :he)
                     (format nil "tries to ~a" 
                             (attack-word (weapon att) :inf)))
                 (name def)
                 (case result (:hit "") (:miss ", but misses") 
                       (:fail ", but doesn't cause any harm"))))



(defmethod combat-situation ((att player) def result)
  (format nil "~@(~a~) ~a ~a~a." 
                 (name att) 
                 (if (eql result :hit) 
                     (attack-word (weapon att) :you)
                     (format nil "try to ~a" 
                             (attack-word (weapon att) :inf)))
                 (name def)
                 (case result (:hit "") (:miss ", but miss") 
                       (:fail ", but do not cause any harm"))))

(defmethod combat-situation (att (def player) result)
  (format nil "~@(~a~) ~a ~a~a." 
                 (name att) 
                 (if (eql result :hit) 
                     (attack-word (weapon att) :he)
                     (format nil "tries to ~a" 
                             (attack-word (weapon att) :inf)))
                 "you"
                 (case result (:hit "") (:miss ", but misses") 
                       (:fail ", but doesn't cause any harm"))))

(defclass e-c-hit (e-combat)
  ((damage :initarg :damage :accessor damage)))

(defmethod description ((event e-c-hit))
  (with-slots ((att attacker) (def defender)) event
    (combat-situation att def :hit)))

(defclass e-c-miss (e-combat) ())

(defmethod description ((event e-c-miss))
  (with-slots ((att attacker) (def defender)) event
    (combat-situation att def :miss)))

(defclass e-c-fail (e-combat) ())

(defmethod description ((event e-c-fail))
  (with-slots ((att attacker) (def defender)) event
    (combat-situation att def :fail)))


(defclass e-monster (event)
  ((monster :initarg :monster :accessor monster)))

(defmethod cast ((event e-monster))
  (loop for m in (remove-duplicates *monsterstack* :key #'car)
        for mon = (car m)
        when (aware-p mon (y (monster event)) (x (monster event)))
        do (react mon event)))

(defclass e-dead (e-monster) ())

(defmethod description ((event e-dead))
  (if (eql (monster event) *player*)
      (display-simple-message "You die...")
      (let ((dead-message 
             (list "~@(~a~) dies.")))
        (format nil (nth (random (length dead-message)) dead-message) 
                (name (monster event))))))

(defclass e-m-angry (e-monster) ())

(defmethod description ((event e-m-angry))
  (format nil "~@(~a~) gets angry." (name (monster event))))

(defclass e-m-calm (e-monster) () )

(defmethod description ((event e-m-calm))
  (format nil "~@(~a~) calms down." (name (monster event))))

(defclass e-poison-start (e-monster) ())

(defmethod description ((event e-poison-start))
  (if (= (effect-count (monster event) :poison) 1)
      (if (eql (monster event) *player*)
          (format nil "You are poisoned.")
          (format nil "~@(~a~) is poisoned." (name (monster event))))
      (if (eql (monster event) *player*)
          (format nil "You are more poisoned.")
          (format nil "~@(~a~) is poisoned, again." (name (monster event))))))

(defclass e-poison-end (e-monster) ())

(defmethod description ((event e-poison-end))
  (if (= (effect-count (monster event) :poison) 1)
      (if (eql (monster event) *player*)
          (format nil "You are no longer poisoned.")
          (format nil "~@(~a~) is no longer poisoned." (name (monster event))))
      (if (eql (monster event) *player*)
          (format nil "You feel better.")
          (format nil "~@(~a~) looks healthier." (name (monster event))))))

(defclass e-poison-damage (e-monster) ())

(defmethod description ((event e-poison-damage))
  (if (eql (monster event) *player*)
      (format nil "You feel worse.")
      (format nil "~@(~a~) doesn't look well." (name (monster event)))))

(defclass e-tar-splash-start (e-monster) ())

(defclass e-tar-splash-end (e-monster) ())

(defmethod description ((event e-tar-splash-start))
  (if (eql (monster event) *player*)
      (format nil "You are splashed with sticky tar!")
      (format nil "~@(~a~) is splashed with sticky tar! Funny."
              (name (monster event)))))

(defmethod description ((event e-tar-splash-end))
  (if (eql (monster event) *player*)
      (format nil "You feel more clean.")
      (format nil "~@(~a~) looks more clean." (name (monster event)))))

(defclass e-acid-splash (e-monster) ())

(defmethod description ((event e-acid-splash))
  (if (eql (monster event) *player*)
      (format nil "Ouch! You are splashed with acid!")
      (format nil "~@(~a~) is splashed with acid!" (name (monster event)))))

(defclass e-item (event)
  ((monster :initarg :monster :accessor monster)
   (item :initarg :item :initform nil :accessor item)))

(defmethod cast ((event e-item))
  (loop for m in (remove-duplicates *monsterstack* :key #'car)
        for mon = (car m)
        when (aware-p mon (y (monster event)) (x (monster event)))
        do (react mon event)))

(defclass e-door (event)
  ((monster :initarg :monster :accessor monster)
   (door :initarg :door :accessor door)))

(defmethod cast ((event e-door))
  (loop for m in (remove-duplicates *monsterstack* :key #'car)
        for mon = (car m)
        when (aware-p mon (y (door event)) (x (door event)))
        do (react mon event)))

(defmethod react ((monster player) (event e-door))
  (declare (ignore monster))
  (maybe-out-char (door event)))

(defclass e-open-door (e-door) ())

(defclass e-close-door (e-door) ())

;;;There should be events for picking up and dropping stuff

;;SECTION 7: Timer and Effects

(defclass timer ()
  ((total-time :initform 0 :accessor total-time)
   (time-hash :initform (make-hash-table) :accessor time-hash)))

(defclass effect ()
  ((monster :initarg :monster :accessor monster)
   (flag :initform :unknown :accessor flag)
   (source :initarg :source :initform nil :accessor source)
   (check-frequency :initarg :check-frequency :initform 5 
                    :accessor check-frequency
                    :documentation "How often to check")))

(defun register (object timer)
  (setf (gethash object (time-hash timer)) (total-time timer)))

(defun unregister (object timer)
  (remhash object (time-hash timer)))

(defun tick (timer time)
  (incf (total-time timer) time)
  (processing timer))

(defun check (object timer)
  (setf (gethash object (time-hash timer)) (total-time timer)))

(defun processing (timer)
  (let ((tt (total-time timer)))
    (maphash (lambda (object time) (process timer (- tt time) object)) 
             (time-hash timer)))) 

(defgeneric process (timer time object) 
  (:documentation "Checks whether the object (which can be an
effect or something else entirely) should do some timed action"))

(defmethod process ((timer timer) time (effect effect))
  (when (> time (check-frequency effect))
    (check effect timer)
    (on-timer effect time))) 

(defgeneric start (effect)
  (:documentation "Called when the effect is applied")
  (:method (effect) (declare (ignore effect)) nil)) 

(defgeneric finish (effect)
  (:documentation "Called when the effect is finished")
  (:method (effect) (declare (ignore effect)) nil))

(defgeneric on-timer (effect time)
  (:documentation "Called when effect is active and time passed
  since the last call")
  (:method (effect time) (declare (ignore effect time)) nil))

(defmethod start :before ((effect effect))
  (register effect *timer*)
  (incf (gethash (flag effect) (effect-hash (monster effect)) 0)))

(defmethod finish :after ((effect effect))
  (decf (gethash (flag effect) (effect-hash (monster effect))))
  (unregister effect *timer*))

(defun effect-count (monster flag)
  (gethash flag (effect-hash monster) 0))

(defclass one-off-effect (effect)
  ((check-frequency :initform 0)))

(defmethod start :before ((effect one-off-effect)) nil)

(defclass permanent-effect (effect)
  ((check-frequency :initform 0)))

(defmethod process ((timer timer) time (effect permanent-effect))
  (declare (ignore time)) nil)

(defclass finite-effect (effect)
  ((duration :initarg :duration :initform 100 :accessor duration)))

(defmethod process ((timer timer) time (effect finite-effect))
  ;;(message "processing effect ~a, time=~a." effect time)
  (when (> time (check-frequency effect))
    (check effect timer)
    (if (< time (duration effect)) 
        (progn (decf (duration effect) time) (on-timer effect time))
        (progn (on-timer effect (duration effect)) (finish effect)))))

(defclass timed-effect (effect)
  ((frequency :initarg :frequency :initform 50.0 :accessor frequency :type float)))

(defmethod on-timer :around ((effect timed-effect) time)
  (when (< (random (frequency effect)) time) (call-next-method))) 

(defclass poison (timed-effect finite-effect)
  ;;parameters: strength, frequency, check-frequency, duration
  ((strength :initarg :strength :initform 1 :accessor strength)
   (flag :initform :poison)
   ))

(defmethod on-timer ((effect poison) time)
  (declare (ignore time))
  (cast (make-instance 'e-poison-damage :monster (monster effect)))
  (decf (hp (monster effect)) (1+ (random (strength effect))))
  (when (<= (hp (monster effect)) 0) 
    (kill (monster effect) :poison "poisoned")
    (when (typep (source effect) 'monster)
      (register-kill (source effect) (monster effect)))))

(defmethod start ((effect poison))
  (cast (make-instance 'e-poison-start :monster (monster effect))))

(defmethod finish ((effect poison))
  (cast (make-instance 'e-poison-end :monster (monster effect))))

(defclass snake-poison (poison) ())

(defclass spider-poison (poison) ())

(defclass tar-splash (finite-effect)
  ((duration :initform 40) (flag :initform :tar)))

(defmethod start ((effect tar-splash))
  (cast (make-instance 'e-tar-splash-start :monster (monster effect)))
  (setf (speed (monster effect)) (* (speed (monster effect)) 5/4)))  
  
(defmethod finish ((effect tar-splash))
  (setf (speed (monster effect)) (* (speed (monster effect)) 4/5))
  (cast (make-instance 'e-tar-splash-end :monster (monster effect))))

(defclass thirst (effect) ())

(defmethod on-timer ((effect thirst) time)
  (decf (thirst (monster effect)) time)
  (let ((th (thirst (monster effect))))
    (cond ((< th 0)
           (message "You are dehydrated...")
           (kill (monster effect) :thirst "dehydrated"))
          ((< 0 (- 150 th) time) (message "You're getting thirsty."))
          ((< 0 (- 50 th) time) (message "You're getting VERY thirsty!")))
    ))

(defclass acid-attack (one-off-effect) 
  ((strength :initarg :strength :initform 3 :accessor strength)))

(defmethod start ((effect acid-attack))
  (cast (make-instance 'e-acid-splash :monster (monster effect)))
  (decf (hp (monster effect)) (1+ (random (strength effect))))
  (when (<= (hp (monster effect)) 0) 
    (kill (monster effect))
    (when (typep (source effect) 'monster)
      (register-kill (source effect) (monster effect)))))
  

;;SECTION 8: Stack processing

(defun occupied (y x)
  (let ((tile (aref *map* y x)))
    (not (and (passable tile) (not (monster tile))))))

(defun random-free-position ()
  (loop for x = (1+ (random 74))
        for y = (1+ (random 19))
        while (occupied y x)
        finally (return (cons y x))))

(defun random-position ()
  (loop for x = (1+ (random 74))
        for y = (1+ (random 19))
        until (passable (aref *map* y x))
        finally (return (cons y x))))

(defun add-to-stack (monster time)
  (let ((first (first *monsterstack*)))
    (when (zerop time)
      (if (eql monster *player*)
          (progn (push (cons monster time) *monsterstack*) 
                 (return-from add-to-stack))
          (setf time 5)))
    (when (or (not first) (> (cdr first) time)) 
      (push (cons monster time) *monsterstack*) 
      (return-from add-to-stack))
    (setf *monsterstack* (nconc *monsterstack* (list (cons monster time))))
    (setq *monsterstack*
          (stable-sort *monsterstack* #'< :key #'cdr))))

(defun spawn (monster &key (position (random-free-position)) (delay 1))
  (add-to-stack monster delay)
  (let* ((x (cdr position)) (y (car position))
         (tile (aref *map* y x)))
    (setf (y monster) y (x monster) x
          (monster tile) monster))
  monster)

(defun move-monster (monster y x)
  (let ((tile (aref *map* y x)))
    (setf (monster (aref *map* (y monster) (x monster))) nil
          (y monster) y (x monster) x
          (monster tile) monster)))

(defun run-ai (monster)
  (setf *level-changed* nil)
  (let ((action (ai monster)))
    (do-action monster action)
    (when (<= *curlevel* 0) (return-from run-ai))
    (let ((time (if *new-time* *new-time* (time action))))
      ;;(format t "Monster ~a does action ~a. Time: ~a. *new-time*=~a. ~%" monster action time *new-time*)
      (when (and (eql monster *player*) 
                 (< (random 100.0) time))
        (generate-monster *curlevel*))
      (when (and (< (random 200.0) time)
                 (< (hp monster) (max-hp monster)))
        (incf (hp monster)))
      (add-to-stack monster (if *cancelled* 0 
                                (* time (speed monster)))))))

(defun run-stack ()
  (unless *monsterstack* (return-from run-stack :quit))
  (let ((min-time (cdar *monsterstack*)))
    (tick *timer* min-time)
    (run-ai (setf *curmonster* (car (pop *monsterstack*))))
    (loop for x in *monsterstack* do (decf (cdr x) min-time))
    (return-from run-stack :ok)))

;;SECTION 9: Generating dungeon

(defun consecutive-zeros (array)
  (loop for x across array
        for n = 0 then (if (zerop x) (1+ n) 0)
        maximizing n))

(defun proper-rooms-p (room1 room2)
  "Checks that rooms don't intersect"
  (destructuring-bind (r1x1 r1y1 r1x2 r1y2) room1
    (destructuring-bind (r2x1 r2y1 r2x2 r2y2) room2
      (or (<= 0 (* (- r1x2 r2x1) (- r1x1 r2x2)))
          (<= 0 (* (- r1y2 r2y1) (- r1y1 r2y2)))))))

(defun random-place (room)
  (destructuring-bind (y1 x1 y2 x2) room
    (cons (+ y1 1 (random (- y2 y1 1)))
          (+ x1 1 (random (- x2 x1 1))))))

(defun tonnelable (premap y1 x1 y2 x2)
  (cond
    ((= y1 y2)
     (when (loop for x from (min x1 x2) to (max x1 x2)
                always (member (aref premap y1 x) '(:h :any :floor)))
       (cons y1 x1)))
    ((= x1 x2)
     (when (loop for y from (min y1 y2) to (max y1 y2)
                always (member (aref premap y x1) '(:v :any :floor)))
       (cons y1 x1)))
    (t (case (random 2)
         (0 (if (and (tonnelable premap y1 x1 y1 x2) 
                     (tonnelable premap y1 x2 y2 x2))
                (cons y1 x2)
                (when (and (tonnelable premap y1 x1 y2 x1) 
                           (tonnelable premap y2 x1 y2 x2))
                  (cons y2 x1))))
         (1 (if (and (tonnelable premap y1 x1 y2 x1) 
                     (tonnelable premap y2 x1 y2 x2))
                (cons y2 x1)
                (when (and (tonnelable premap y1 x1 y1 x2)
                           (tonnelable premap y1 x2 y2 x2))
                  (cons y1 x2))))))))

(defun generate-map (&key stairs player (delay 0) use-existing-player
                     &aux upstair downstair)
  (setf *map* (make-array '(21 76)))
  (setf *known-map* (make-array '(21 76) :element-type 'bit))
  (setf *monsterstack* nil *roompoints* nil *rooms* nil)
  ;(loop for x from 0 to 75
  ;      do (loop for y from 0 to 20
  ;              do (setf (aref *map* y x) 
  ;                       (make-instance 'wall :x x :y y))))
  (let ((rooms nil) ;;each room is coded as (y1 x1 y2 x2)
        (premap (make-array '(21 76))))
    (loop for x from 0 to 75
          do (loop for y from 0 to 20
                   do (setf (aref premap y x) :any)))
    ;Make rooms
    (loop with free-cols = (make-array 76 :element-type 'bit 
                                       :initial-element 0)
          with free-rows = (make-array 21 :element-type 'bit
                                       :initial-element 0)
          for ncols = (consecutive-zeros free-cols)
          for nrows = (consecutive-zeros free-rows)
          while (or (>= ncols 8) (>= nrows 5)) 
          do (loop for width = ;(if (< nrows 5) (min ncols (+ (random 8) 6))
                                   (+ (random 8) 6);)
                   for height = ;(if (< ncols 8) (min nrows (+ (random 4) 5))
                                    (+ (random 4) 5);)
                   for start-x = (random (- 74 width))
                   for start-y = (random (- 19 height))
                   for room = (list start-y start-x
                                    (+ start-y height)
                                    (+ start-x width))
                   until (every (lambda (x) (proper-rooms-p room x))
                                rooms)
                  finally (push room rooms))
             (let ((newroom (first rooms)))
               (destructuring-bind (y1 x1 y2 x2) newroom
                 (loop for i from y1 to y2
                       do (setf (bit free-rows i) 1))
                 (loop for i from x1 to x2
                       do (setf (bit free-cols i) 1))
                 ;Dig room
                 (loop for y from y1 to y2
                       do (loop for x from x1 to x2
                                for wallx = (* (- x1 x) (- x2 x))
                                for wally = (* (- y1 y) (- y2 y))
                                do (setf (aref premap y x)
                                         (cond
                                           ((= 0 wallx wally) :none)
                                           ((zerop wallx) :h)
                                           ((zerop wally) :v)
                                           (t :floor))))))))
    #-(and) 
    (loop for y from 0 to 20
       do (loop for x from 0 to 75
             do (princ (case (aref premap y x)
                         (:any #\*)
                         (:none #\#)
                         (:h #\|)
                         (:v #\-)
                         (:floor #\.))))
         (terpri))
    ;(format t "Rooms are generated.~%")
    ;(finish-output)
    ;Make tonnels
    (loop with connected-rooms = (list (first rooms))
          for i from 1 to (1- (length rooms))
          for other-rooms = (set-difference rooms connected-rooms
                                            :test #'equal)
          do (loop for room1 = (random-list connected-rooms)
                   for room2 = (random-list other-rooms)
                   for (y1 . x1) = (random-place room1)
                   for (y2 . x2) = (random-place room2)
                   for tonnel = (tonnelable premap y1 x1 y2 x2)
                   until tonnel
                   finally ;;Dig tonnel
                   (labels ((cornerify-h (y x)
                              (case (aref premap y x)
                                ((:any :v) (setf (aref premap y x) :none))))
                            (cornerify-v (y x)
                              (case (aref premap y x)
                                ((:any :h) (setf (aref premap y x) :none))))
                            (wallify-h (y x)
                              (case (aref premap y x)
                                ((:noney :any) (setf (aref premap y x) :v))
                                (:h (setf (aref premap y x) :none))))
                            (wallify-v (y x)
                              (case (aref premap y x)
                                ((:noney :any) (setf (aref premap y x) :h))
                                (:v (setf (aref premap y x) :none))))
                            (tonnel (y1 x1 y2 x2)
                              "Digs straight tonnel between 2 points."
                              (let ((xmin (min x1 x2)) (xmax (max x1 x2))
                                    (ymin (min y1 y2)) (ymax (max y1 y2)))
                                (cond
                                  ((and (= x1 x2) (= y1 y2))
                                   (return-from tonnel))
                                  ((= y1 y2)
                                   ;Corners
                                   (cornerify-h (1- y1) (1- xmin))
                                   (cornerify-h (1+ y1) (1- xmin))
                                   (cornerify-h (1- y1) (1+ xmax))
                                   (cornerify-h (1+ y1) (1+ xmax))
                                   (loop for x from xmin to xmax
                                      do (setf (aref premap y1 x) :floor)
                                         (wallify-h (1- y1) x)
                                         (wallify-h (1+ y1) x)))
                                  ((= x1 x2)
                                   ;Corners
                                   (cornerify-v (1- ymin) (1- x1))
                                   (cornerify-v (1- ymin) (1+ x1))
                                   (cornerify-v (1+ ymax) (1- x1))
                                   (cornerify-v (1+ ymax) (1+ x1))
                                   (loop for y from ymin to ymax
                                        do (setf (aref premap y x1) :floor)
                                           (wallify-v y (1- x1))
                                           (wallify-v y (1+ x1))))))))
                     (destructuring-bind (y3 . x3) tonnel
                       (tonnel y1 x1 y3 x3)
                       (tonnel y3 x3 y2 x2))
                     (push room2 connected-rooms))))
    #-(and)
    (loop for y from 0 to 20
          do (loop for x from 0 to 75
                do (princ (case (aref premap y x)
                            (:any #\*)
                            (:none #\#)
                            (:h #\|)
                            (:v #\-)
                            (:floor #\.))))
             (terpri))
    (loop for y from 0 to 20
          do (loop for x from 0 to 75
                   
                   do (setf (aref *map* y x) 
                            (if (eql (aref premap y x) :floor)
                                (make-instance 'concrete-floor :x x :y y)
                                (make-instance 'wall :x x :y y)))))
    (loop for r in rooms do (push (random-place r) *roompoints*))
    (setf *rooms* rooms)
    )

;;Doors

  (dolist (r *rooms*)
    (destructuring-bind (y1 x1 y2 x2) r
      (loop with stepx = 1 and stepy = 0
         for x = (1+ x1)  then (+ x stepx)
         for y = y1 then (+ y stepy)
         until (and (= x1 x) (= y1 y))
         do (cond
              ((and (= x2 x) (= y1 y))
               (setf stepx 0) (setf stepy 1))
              ((and (= x2 x) (= y2 y))
               (setf stepx -1) (setf stepy 0))
              ((and (= x1 x) (= y2 y))
               (setf stepx 0) (setf stepy -1))
              (t (when (and (typep (aref *map* y x) 'concrete-floor) (zerop (random 3)))
                   (setf (aref *map* y x) (make-instance 'door :x x :y y))))))))

;;Staircases

  (ecase stairs
    ((nil) nil)
    (:u (destructuring-bind (y . x) (first *roompoints*)
          (setf (aref *map* y x) 
                (make-instance 'staircase :direction :u :y y :x x))
          (setf upstair (cons y x))))
    (:d (destructuring-bind (y . x) (second *roompoints*)
          (setf (aref *map* y x) 
                (make-instance 'staircase :direction :d :y y :x x))
          (setf downstair (cons y x))))
    (:ud (destructuring-bind (y . x) (first *roompoints*)
           (setf (aref *map* y x) 
                 (make-instance 'staircase :direction :u :y y :x x))
           (setf upstair (cons y x))) 
         (destructuring-bind (y . x) (second *roompoints*)
           (setf (aref *map* y x) 
                 (make-instance 'staircase :direction :d :y y :x x))
           (setf downstair (cons y x)))))

;;Spawning player
  (if use-existing-player
      (ecase player
            (:u (move-monster *player* (car *upstair*) (cdr *upstair*)))
            (:d (move-monster *player* (car *downstair*) (cdr *downstair*)))
            (:random (destructuring-bind (y . x) (random-free-position)
                       (move-monster *player* y x))))        
      (setf *player*
            (ecase player
              ((nil) nil)
              (:u (spawn (make-instance 'player) :position upstair :delay delay))
              (:d (spawn (make-instance 'player) :position downstair :delay delay))
              (:random (spawn (make-instance 'player) :delay delay)))))
  
  (update-fov *player*)
;;Spawning monsters

  (generate-monsters *curlevel*)

;;Spawning items

  (generate-items *curlevel*)

)
                 
;;SECTION 10: Printing routines


(defun out-char (obj &key debug &aux clr)
  "Actually outputs the character of tile, monster or item"
  (multiple-value-bind (char color) (print-char obj)
    (setf clr color)
    (when (zerop (bit (fov *player*) (y obj) (x obj)))
      ;;(when (typep obj 'door) (return-from out-char))
      (setf clr :cgray))
    (when (zerop (bit *known-map* (y obj) (x obj)))
      (return-from out-char))
    (when (typep obj 'door) (setf clr color)) 
    (when debug (setf clr (if (eql debug t) :cred debug)))
    (attrset clr)
    (mvaddch (+ *dy* (y obj)) (+ *dx* (x obj)) char)))

(defun maybe-out-char (obj)
  "Outputs the character if it's inside players' field of vision"
  (unless (zerop (bit (fov *player*) (y obj) (x obj)))
    (out-char obj)))

(defun print-map ()
  (loop for x from 0 to 75
        do (loop for y from 0 to 20
                do (out-char (aref *map* y x)))))

(defun clean-line (line)
  (loop for i from 0 to 79 do (mvaddch line i #\Space)))

(defun clean-area (y1 x1 y2 x2)
  (loop for y from y1 to y2
        do (loop for x from x1 to x2
                 do (mvaddch y x #\Space))))

(defun safe-print (y x str)
  "Workaround for a bug"
  (loop for c across str
        for xx = x then (1+ xx)
        do (mvaddch y xx c))) 

(defun display-message (str)
  (if str 
      (progn
        (attrset :cdark)
        (clean-line 0)
        (safe-print 0 1 str)
        (when *messages*
          (attrset :cyellow)
          (mvprintw 0 74 "[MORE]"))
        (refresh))
      (display-simple-message "No more messages.")))

(defun display-simple-message (str)
  (attrset :cdark)
  (clean-line 0)
  (safe-print 0 1 str)
  (push str *allmessages*)
  (refresh) "")

(defun message (str &rest args)
  (let ((s (apply #'format nil str args)))
    (push s *messages*) (push s *allmessages*)))

(defun merge-messages ()
  (loop with mstr = ""
        for m in *messages*
        if (<= (+ (length m) (length mstr)) 72)
        do (setf mstr (concatenate 'string mstr m " "))
        else collect mstr into mess
             and do (setf mstr (concatenate 'string m " "))
        finally (setf *messages* (nconc mess (list mstr)))))

(defun show-items (y x)
  (clean-line 23)
  (let ((items (items (aref *map* y x)))
        (curpos 1))
    (loop for i in items
          for n = (description i)
          do (attrset (nth-value 1 (print-char i)))
          if (< (+ curpos (length n)) 80)
             do (mvprintw 23 curpos n)
                (incf curpos (+ (length n) 2))
          else do (let ((diff (- 79 curpos)))
                    (when (>= diff 0)
                      (mvprintw 23 curpos (subseq n 0 diff)))
                    (mvprintw 23 77 "...")
                    (loop-finish)))))

(defun show-status ()
  (clean-line 22)
  (attrset :CDARK)
  (mvprintw 22 2 "HP:")
  (when (> (effect-count *player* :poison) 0) (attrset :cdgreen))
  (printw (format nil "~d" (hp *player*)))
  (attrset :Cdark)
  (printw (format nil "/~d  STR:~d  DEX:~d  AWA:~d" 
                  (max-hp *player*) 
                  (strength *player*)
                  (dexterity *player*)
                  (awareness *player*)))
  (when (> (effect-count *player* :tar) 0)
    (attrset :ccyan) (printw " Slow"))
  (cond ((< (thirst *player*) 50)
         (attrset :crose) (printw " Thirsty!"))
        ((< (thirst *player*) 150)
         (attrset :cyellow) (printw " Thirsty")))
  (attrset :Cdark)
  (mvprintw 22 50 (format nil "Level:~d" *curlevel*))
  (attrset :cyellow)
  (mvprintw 22 60 (format nil "Cash: $~d" (cash *player*))))

(defun redraw-screen ()
  (erase)
  (print-map)
  (show-status)
  (show-items (y *player*) (x *player*))
  (move (+ *dy* (y *player*)) (+ *dx* (x *player*)))
  (refresh))

(defun recap-messages ()
  (erase)
  (attrset :cgray)
  (loop for x = *allmessages* then (cdr x)
     for i from 20 downto 1
     do (safe-print i 1 (if (car x) (car x) "-"))))

(defun yes-or-no (str)
  (display-simple-message (format nil "~a (y/n)" str))
  (case (getch)
    ((121 173 89 141) t)
    ((110 78 226 146) nil)))

(defun choose-direction ()
  (display-simple-message "Choose direction [12346789]")
  (case (getch)
    ((261 54 454) (cons 0 1))
    ((260 52 452) (cons 0 -1))
    ((259 56 450) (cons -1 0))
    ((258 50 456) (cons 1 0))
    ((51 457) (cons 1 1))
    ((49 455) (cons 1 -1))
    ((57 451) (cons -1 1))
    ((55 449) (cons -1 -1))
    (t nil)))

;;SECTION 11: Serialization module

(defun save-game (file)
  (cl-store:store (list *allmaps* *curlevel* *deepest-level*
                        *player* *allmessages* *timer* *map* *known-map* *roompoints*
                        *rooms* *monsterstack*) file))

(defun load-game (file)
  (setf (values *allmaps* *curlevel* *deepest-level* *player* *allmessages* *timer*
                *map* *known-map* *roompoints* *rooms* *monsterstack*)
        (values-list (cl-store:restore file)))
  (remove *player* *monsterstack* :key #'car)
  (push (cons *player* 0) *monsterstack*))

(defun go-to-level (n &key (player :u) newgame)
  (let ((target-level (aref *allmaps* (1- n))))
    ;;;Saving old map
    (unless newgame
      (setf (aref *allmaps* (1- *curlevel*))
            (list *map* *known-map* *roompoints* *rooms* *monsterstack*))
      (dotimes (i 21)
            (dotimes (j 76)
              (setf (bit (fov *player*) i j) 0))))
    ;;;Loading new map
    (setf *curlevel* n)
    (when (> *curlevel* *deepest-level*) (setf *deepest-level* *curlevel*))
    (if (and target-level (not newgame))
        (progn
          (setf (values *map* *known-map* *roompoints* *rooms* *monsterstack*)
                (values-list target-level))
          (remove *player* *monsterstack* :key #'car)
          (ecase player
            (:u (move-monster *player* (car *upstair*) (cdr *upstair*)))
            (:d (move-monster *player* (car *downstair*) (cdr *downstair*)))
            (:random (destructuring-bind (y . x) (random-free-position)
                       (move-monster *player* y x))))
          ;(add-to-stack *player* 5))
          )
        (progn (setf *monsterstack* nil) 
               (generate-map :stairs (if (< *curlevel* 10) :ud :u) 
                             :player player :delay (if newgame 0 5)
                             :use-existing-player (not newgame))
               (setf (aref *allmaps* (1- *curlevel*))
                     (list *map* *known-map* *roompoints* *rooms* *monsterstack*))))
    (update-fov *player*)
    (redraw-screen)
    (setf *level-changed* t)))

    
;;SECTION 12: Main sequence

(defun splash-screen ()
  (attrset :cgreen)
  (mvprintw 1 1 "The Sewer Massacre 1.0")
  (attrset :cwhite)
  (mvprintw 2 1 "(c) 2006, 2007 Timofei Shatrov")
  (attrset :cyellow)
  (mvprintw 3 1 "Originally written for 2006 Orange October Minigun 7DRL Contest")
  (attrset :crose)
  (mvprintw 4 10 "And later drastically improved! ")
  (attrset :cwhite)
  (mvprintw 24 50 "[press any key to continue]")
  (refresh)
  (getch))

(defun genmd5 (object)
  (coerce (md5sum-sequence (let ((*print-readably* t)) (format nil "~s" object)))
          'list))

(defun load-hiscore ()
  (unless (probe-file "hiscore") (return-from load-hiscore nil))
  (with-open-file (in "hiscore"
                      :external-format #+clisp charset:utf-8 #+sbcl :utf-8)
    (let* ((*read-eval* nil)
           (hashentry (read in))
           (hiscore (read in)))
      (if (equal (genmd5 hiscore) hashentry) hiscore :error))))

(defun save-hiscore (hiscore)      
  (with-open-file (out "hiscore" :direction :output :if-exists :supersede
                       :external-format #+clisp charset:utf-8 #+sbcl :utf-8)
    (let ((*print-readably* t))
      (format out "~s~%~s " (genmd5 hiscore) hiscore))))

(defun status (flag)
  (case flag
    (:kill (format nil "[killed by ~a on L~d]" *quitreason* *curlevel*))
    (:quit "[missing]")
    (:poison (format nil "[killed by poison on L~d]" *curlevel*))
    (:thirst (format nil "[dehydrated on L~d]" *curlevel*))
    (t "[alive]")))

(defun calc-score (player)
  (* (cash player)
     (case *quitflag*
       (:left 2)
       (t 1))))

(defun hiscore (&aux hiscore (number 0))
  (setf hiscore (load-hiscore))
  (when (eql hiscore :error)
    (erase)
    (attrset :cred)
    (mvprintw 1 1 "Your hiscore file is corrupted. Please delete it.")
    (attrset :cwhite)
    (mvprintw 24 50 "[press any key to quit]")
    (getch)
    (return-from hiscore))
    
  (setf hiscore (nreverse hiscore))
  (setf hiscore (stable-sort (nreverse (push (cons :player 
                                                   (cons (calc-score *player*)
                                                         (status *quitflag*)))
                                             hiscore))
                             #'> :key #'cadr))
  (when (> (length hiscore) *hiscore-length*)
    (setf hiscore (nbutlast hiscore)))
  (let ((ask (loop for entry in hiscore
                   do (incf number)
                   when (eql (car entry) :player) do (loop-finish)
                   finally (if (eql (car entry) :player) (return entry) nil))))
             ;(find :player hiscore :key #'car)))
    (when ask
      (erase) (attrset :cyellow)
      (mvprintw 1 1 "Congratulations! You made the top ten!")
      (attrset :grey) (mvprintw 2 1 "Enter your name: ")
      (setf (car ask) (getnstr 12)))
    (unless ask (setf number 0)))

  (erase)
  (attrset :cgreen)
  (mvprintw 1 1 (format nil "Top ~d Adventurers:" *hiscore-length*))
  (attrset :cwhite)
  (loop for i from 1 to *hiscore-length*
        for entry in hiscore
        when (= i number) do (attrset :cyellow)
        do (mvprintw (+ 2 i) 1 (format nil "~2d. ~15a~10a ~a" 
                                       i (car entry) (cadr entry)
                                       (cddr entry)))
        when (= i number) do (attrset :cwhite))
  (save-hiscore hiscore)

  (mvprintw 24 50 "[press any key to quit]")
  (refresh)
  (getch))

(defun morgue-dump ()
  (with-open-file (out "morgue.txt" :direction :output :if-exists :supersede)
    (format out "The Sewer Massacre 1.0 character dump~%")
    (format out "----------------------------------------~%")
    (format out "HP:~d/~d  STR:~d  DEX:~d  AWA:~d~%" 
                         (hp *player*) (max-hp *player*) 
                         (strength *player*)
                         (dexterity *player*)
                         (awareness *player*))
    (format out "Level:~d   Deepest level visited:~d   Cash: $~d~%" 
            *curlevel* *deepest-level* (cash *player*))
    (format out "Status: ~a~%" (status *quitflag*))
    (format out "Score: ~d~%" (calc-score *player*))
    (terpri out)
    (format out "Inventory~%")
    (format out "---------~%")
    (loop for item in (inventory *player*)
          do (format out "~a~%" (if (equipped item)
                                    (format nil "~a [equipped]"
                                            (description item))
                                    (description item))))
    (terpri out)
    (format out "Last messages~%")
    (format out "-------------~%")
    (loop for mess in *allmessages*
          for i from 1 to 20
          collecting mess into bucket
          finally (dolist (m (reverse bucket)) (format out "~a~%" m)))
    
    (terpri out)
    (format out "Kills~%")
    (format out "-----~%")
    (when (> (kill-count *player*) 0)
      (loop for mon being each hash-key in (kill-hash *player*) using (hash-value n)
         do (format out "~a: ~d~%" mon n)))
    (terpri out)
    (format out "Total kills: ~d~%" (kill-count *player*))
    ))

(defun runtime ()
  (setf *quitflag* nil)
  (loop 
     (run-stack) 
     (when *quitflag*
       (unless (eql *quitflag* :saved)
         (when (probe-file "current.sav")
             (delete-file "current.sav"))
         (hiscore)
         (morgue-dump))
       ;(hiscore)
       (return))))

(defun start-game (&key debug)
  (setf *random-state* (make-random-state t)
        *allmessages* nil
        *messages* nil
        *timer* (make-instance 'timer)
        *monster-generator* (make-array 10 :initial-element nil)
        *item-generator* (make-array 10 :initial-element nil)
        *allmaps* (make-array 10 :initial-element nil)
        *debug-mode* debug
        *quitreason* ""
        *deepest-level* 1)
  (if debug (connect-console)
   (create-console "textconsole.exe"))
  (erase)
  (splash-screen)
  (unwind-protect
       (progn
         (when (init-controls) (error "Cannot find control file. This sucks :("))
         (erase)
         (if (probe-file "current.sav")
             (load-game "current.sav")
             (progn
               (setf *curlevel* 1) (go-to-level 1 :newgame t)))
         (print-map)
         (refresh)
         (runtime))
    (close-console))
  (unless debug 
    #+clisp (ext:quit)
    #+sbcl (sb-ext:quit)))


#+clisp
(defun make-exec ()
  (ext:saveinitmem "sewers.exe" 
                   :quiet t :norc t :init-function #'start-game 
                   :start-package :sewers :executable t))
